#pragma once

#include <GL/glew.h>
#include <math.h>
#include "linmath.h"

inline GLfloat radians(GLfloat deg)
{
    return deg/180.f* 3.141592653589793;
}

typedef enum {
	FORWARD,
	BACKWARD,
	LEFT,
	RIGHT
} Camera_Movement;

typedef struct Camera Camera;

static void GetViewMatrix(Camera *camera, mat4x4 view);
static void updateCameraVectors(Camera *camera);
static void ProcessKeyboard(Camera *camera, Camera_Movement direction, GLfloat deltaTime);
static void ProcessMouseMovement(Camera *camera, GLfloat xoffset, GLfloat yoffset, GLboolean constrainPitch /* = true */);
static void ProcessMouseScroll(Camera *camera, GLfloat yoffset);

// Default camera values
static const GLfloat YAW = -90.0f;
static const GLfloat PITCH = 0.0f;
static const GLfloat SPEED = 3.0f;
static const GLfloat SENSITIVTY = 0.25f;
static const GLfloat ZOOM = 45.0f;

typedef struct Camera {
    // Camera Attributes
    vec3 Position;
    vec3 Front;
    vec3 Up;
    vec3 Right;
    vec3 WorldUp;
    // Eular Angles
	GLfloat Yaw;
	GLfloat Pitch;
	// Camera options
	GLfloat MovementSpeed;
	GLfloat MouseSensitivity;
	GLfloat Zoom;
	// Camera methods
    void (*GetViewMatrix)(Camera *camera, mat4x4 view);
    void (*ProcessKeyboard)(Camera *camera, Camera_Movement direction, GLfloat deltaTime);
    void (*ProcessMouseMovement)(Camera *camera, GLfloat xoffset, GLfloat yoffset, GLboolean constrainPitch);
    void (*ProcessMouseScroll)(Camera *camera, GLfloat yoffset);
} Camera;

Camera camera3f(GLfloat posX, GLfloat posY, GLfloat posZ) {
    Camera camera;
    camera.Position[0] = posX;
    camera.Position[1] = posY;
    camera.Position[2] = posZ;
    camera.WorldUp[0] = 0.f;
    camera.WorldUp[1] = 1.f;
    camera.WorldUp[2] = 0.f;
    camera.Yaw = YAW;
    camera.Pitch = PITCH;
    camera.Front[0] = 0.f;
    camera.Front[1] = 0.f;
    camera.Front[2] = -1.f;
    camera.MovementSpeed = SPEED;
    camera.MouseSensitivity = SENSITIVTY;
    camera.Zoom = ZOOM;
    updateCameraVectors(&camera);
    camera.GetViewMatrix = GetViewMatrix;
    camera.ProcessKeyboard = ProcessKeyboard;
    camera.ProcessMouseMovement = ProcessMouseMovement;
    camera.ProcessMouseScroll = ProcessMouseScroll;
    return camera;
}

static void updateCameraVectors(Camera *camera)
{
		// Calculate the new Front vector
		vec3 front;
		front[0] = cos(radians(camera->Yaw)) * cos(radians(camera->Pitch));
		front[1] = sin(radians(camera->Pitch));
		front[2] = sin(radians(camera->Yaw)) * cos(radians(camera->Pitch));
		vec3_norm(camera->Front, front);

		// Also re-calculate the Right and Up vector
		vec3 cross;
		vec3_mul_cross(cross, camera->Front, camera->WorldUp);
		vec3_norm(camera->Right, cross);  // Normalize the vectors, because their length gets closer to 0 the more you look up or down which results in slower movement.
		vec3_mul_cross(cross, camera->Right, camera->Front);
		vec3_norm(camera->Up, cross);
}


static void GetViewMatrix(Camera *camera, mat4x4 view)
{
    vec3 center;
    vec3_add(center, camera->Position, camera->Front);
    mat4x4_look_at(view, camera->Position, center, camera->Up);
}

// Processes input received from any keyboard-like input system. Accepts input parameter in the form of camera defined ENUM (to abstract it from windowing systems)
static void ProcessKeyboard(Camera *camera, Camera_Movement direction, GLfloat deltaTime)
{
	GLfloat velocity = camera->MovementSpeed * deltaTime;
	vec3 scale;
	if (direction == FORWARD) {
        vec3_scale(scale, camera->Front, velocity);
        vec3_add(camera->Position, camera->Position, scale);
    }
	if (direction == BACKWARD) {
		vec3_scale(scale, camera->Front, velocity);
        vec3_sub(camera->Position, camera->Position, scale);
	}
	if (direction == LEFT) {
		vec3_scale(scale, camera->Right, velocity);
        vec3_sub(camera->Position, camera->Position, scale);
	}
	if (direction == RIGHT) {
		vec3_scale(scale, camera->Right, velocity);
        vec3_add(camera->Position, camera->Position, scale);
	}
}

// Processes input received from a mouse input system. Expects the offset value in both the x and y direction.
static void ProcessMouseMovement(Camera *camera, GLfloat xoffset, GLfloat yoffset, GLboolean constrainPitch /* = true */)
{
	xoffset *= camera->MouseSensitivity;
	yoffset *= camera->MouseSensitivity;

	camera->Yaw += xoffset;
	camera->Pitch += yoffset;

	// Make sure that when pitch is out of bounds, screen doesn't get flipped
	if (constrainPitch)
	{
		if (camera->Pitch > 89.0f)
			camera->Pitch = 89.0f;
		if (camera->Pitch < -89.0f)
			camera->Pitch = -89.0f;
	}
	// Update Front, Right and Up Vectors using the updated Eular angles
	updateCameraVectors(camera);
}

// Processes input received from a mouse scroll-wheel event. Only requires input on the vertical wheel-axis
static void ProcessMouseScroll(Camera *camera, GLfloat yoffset)
{
	if (camera->Zoom >= 1.0f && camera->Zoom <= 90.0f)
		camera->Zoom -= yoffset;
	if (camera->Zoom <= 1.0f)
		camera->Zoom = 1.0f;
	if (camera->Zoom >= 90.0f)
		camera->Zoom = 90.0f;
}
