/*------------------------------------------------*/
/* 10assimp/assimp/model.c                         */
#define _CRT_SECURE_NO_WARNINGS
#include <stdbool.h>
#include <string.h> //dirname
#include "model.h"
#include "mesh.h"
#include "linmath.h"

static void loadModel(Model * model, char *path);
static size_t getMeshesCount(const struct aiNode const *  scene);
static void processNode(Model *model, struct aiNode* node, const struct aiScene* scene);
static void drawModel(Model * model, mat4x4 uMVMatrix, mat4x4 view, mat4x4 uPMatrix);
static size_t getIndicesCount(const struct aiMesh * aiMesh);
static Mesh processMesh(Model * model, const struct aiMesh* aiMesh, const struct aiScene* scene);
static void loadMaterialTextures(struct aiMaterial * material, enum aiTextureType type, Texture *textures);
static GLuint textureFromFile(char * path);
static void destroyModel(Model * model);


static size_t meshIndex;
static char directory[MAXLEN];

Model constructModel(char * path)
{
	Model model;
	model.success = true;
	model.draw = drawModel;
	model.destroy = destroyModel;
	model.program = shader("model.vs", "model.fs");
	if (!model.program.Program) {
		model.success = false;
		return model;
	}
	loadModel(&model, path);
	return model;
}

static void loadModel(Model * model, char *path)
{
	const struct aiScene* scene = aiImportFile(path, aiProcess_FlipUVs | aiProcess_Triangulate);
	if (!scene || scene->mFlags == AI_SCENE_FLAGS_INCOMPLETE || !scene->mRootNode) {
		fprintf(stderr, "aiImportFile error: %s\n", aiGetErrorString());
		model->success = false;
		return;
	}
	model->numMeshes = getMeshesCount(scene->mRootNode);
	model->meshes = malloc(sizeof(Mesh) * model->numMeshes);

	strncpy(directory, path, MAXLEN);
	char * sep = strrchr(directory, '/');
	*(sep + 1) = '\0';
	printf("Directory:\t\t\t%s\n", directory);

	meshIndex = 0;
	printf("Count of meshes/lights:\t\t%lu %i\n", model->numMeshes, scene->mNumLights);
	printf("Node processing:\t\tstart\n");
	processNode(model, scene->mRootNode, scene);
	printf("Node processing:\t\tdone\n");

	aiReleaseImport(scene);
}

static size_t getMeshesCount(const struct aiNode const *  node) {
	size_t count = node->mNumMeshes;
	for (GLuint i = 0; i < node->mNumChildren; i++)
	{
		count += getMeshesCount(node->mChildren[i]);
	}
	return count;
}

static void processNode(Model *model, struct aiNode* node, const struct aiScene* scene)
{
	const struct aiMesh* aiMesh;
	static unsigned mashCounter = 0;
	Mesh mesh;

	for (GLuint i = 0; i < node->mNumMeshes; i++)
	{
		aiMesh = scene->mMeshes[node->mMeshes[i]];
		printf("Processing mesh num.:\t\t%u\n", mashCounter++);
		mesh = processMesh(model, aiMesh, scene);
		model->meshes[meshIndex++] = mesh;
	}
	for (GLuint i = 0; i < node->mNumChildren; i++) 	{
		processNode(model, node->mChildren[i], scene);
	}
}

static Mesh processMesh(Model * model, const struct aiMesh* aiMesh, const struct aiScene* scene)
{
	unsigned int i;
	Mesh mesh = constructMesh();
	// 1. vertices
	mesh.vertices = malloc(sizeof(Vertex)* aiMesh->mNumVertices);
	mesh.numVertices = aiMesh->mNumVertices;

	for (i = 0; i < aiMesh->mNumVertices; i++)
	{
		Vertex vertex;
		vertex.Position[0] = aiMesh->mVertices[i].x;
		vertex.Position[1] = aiMesh->mVertices[i].y;
		vertex.Position[2] = aiMesh->mVertices[i].z;
		if (aiMesh->mNormals != NULL) {
			vertex.Normal[0] = aiMesh->mNormals[i].x;
			vertex.Normal[1] = aiMesh->mNormals[i].y;
			vertex.Normal[2] = aiMesh->mNormals[i].z;
		} else {
			fprintf(stderr,"No normals found!\n");
			vertex.Normal[0] = 0.0;
			vertex.Normal[1] = 0.0;
			vertex.Normal[2] = 1.0;
		}

		if (aiMesh->mTextureCoords[0])
		{
			vertex.TexCoords[0] = aiMesh->mTextureCoords[0][i].x;
			vertex.TexCoords[1] = aiMesh->mTextureCoords[0][i].y;
		}
		mesh.vertices[i] = vertex;
	}
	// 2. indices
	mesh.numIndices = getIndicesCount(aiMesh);
	mesh.indices = malloc(sizeof(GLuint) * mesh.numIndices);
	size_t ix = 0;
	for (i = 0; i < aiMesh->mNumFaces; i++)
	{
		struct aiFace face = aiMesh->mFaces[i];
		for (GLuint j = 0; j < face.mNumIndices; j++) {
			mesh.indices[ix++] = face.mIndices[j];
		}
	}

	// 3. materilal
	struct aiMaterial * material = scene->mMaterials[aiMesh->mMaterialIndex];

	// 3.1 colors
	const struct aiColor4D _def_color = {0.0, 0.0, 0.0, 0.0};
	struct aiColor4D _color = {0.0, 0.0, 0.0, 1.0};
	if (AI_SUCCESS == aiGetMaterialColor(material, AI_MATKEY_COLOR_AMBIENT, &_color)) {
		//printf("ambient = {%.2f, %.2f, %.2f, %.2f}, ",_color.r,_color.g,_color.b,_color.a);
	}
	mesh.ambientColor[0] = _color.r;
	mesh.ambientColor[1] = _color.g;
	mesh.ambientColor[2] = _color.b;
	mesh.ambientColor[3] = _color.a;

	_color = _def_color;
	if (AI_SUCCESS == aiGetMaterialColor(material, AI_MATKEY_COLOR_DIFFUSE, &_color)) {
		//printf("diffuse = {%.2f, %.2f, %.2f, %.2f}, ",_color.r,_color.g,_color.b,_color.a);
	}
	mesh.diffuseColor[0] = _color.r;
	mesh.diffuseColor[1] = _color.g;
	mesh.diffuseColor[2] = _color.b;
	mesh.diffuseColor[3] = _color.a;

	_color = _def_color;
	if (AI_SUCCESS == aiGetMaterialColor(material, AI_MATKEY_COLOR_SPECULAR, &_color)) {
		//printf("specular = {%.2f, %.2f, %.2f, %.2f}, ", _color.r, _color.g, _color.b, _color.a);
	}
	mesh.specularColor[0] = _color.r;
	mesh.specularColor[1] = _color.g;
	mesh.specularColor[2] = _color.b;
	mesh.specularColor[3] = _color.a;

	_color = _def_color;
	if (AI_SUCCESS == aiGetMaterialColor(material, AI_MATKEY_COLOR_EMISSIVE, &_color)) {
		//printf("emissive = {%.2f, %.2f, %.2f, %.2f} ", _color.r, _color.g, _color.b, _color.a);
	}
	mesh.emissiveColor[0] = _color.r;
	mesh.emissiveColor[1] = _color.g;
	mesh.emissiveColor[2] = _color.b;
	mesh.emissiveColor[3] = _color.a;
	//printf("\n");

	float opacity;
	if (AI_SUCCESS == aiGetMaterialFloatArray(material, AI_MATKEY_OPACITY, &opacity, NULL)) {
		mesh.ambientColor[3] = opacity;
		mesh.diffuseColor[3] = opacity;
		mesh.specularColor[3] = opacity;
	}
	mesh.shininess = 0.0f;
	aiGetMaterialFloatArray(material, AI_MATKEY_SHININESS, &(mesh.shininess), NULL);

	// 3.2 textures
	mesh.ambientTexCount = aiGetMaterialTextureCount(material, aiTextureType_AMBIENT);
	mesh.diffuseTexCount = aiGetMaterialTextureCount(material, aiTextureType_DIFFUSE);
	mesh.specularTexCount = aiGetMaterialTextureCount(material, aiTextureType_SPECULAR);
	mesh.numTextures = 3;
	mesh.textures = malloc(sizeof(Texture) * mesh.numTextures);

	loadMaterialTextures(material, aiTextureType_AMBIENT, &(mesh.textures[0]));
	loadMaterialTextures(material, aiTextureType_DIFFUSE, &(mesh.textures[1]));
	loadMaterialTextures(material, aiTextureType_SPECULAR, &(mesh.textures[2]));

	setupMesh(&mesh);
	return mesh;
}

static void loadMaterialTextures(struct aiMaterial * material, enum aiTextureType type, Texture *textures)
{
	struct aiString str;
	char path[MAXLEN];
	Texture texture;
	unsigned int texCount = aiGetMaterialTextureCount(material, type);
	if (!texCount) {
		texture.id = 0;
		texture.type = aiTextureType_NONE;
		texture.path = NULL;
		*textures = texture;
		return;
	}
	aiGetMaterialTexture(material, type, 0, &str, NULL, NULL, NULL, NULL, NULL, NULL);
	strncpy(path, directory, MAXLEN);
	strncat(path, str.data, MAXLEN);
	printf("Load texture:\t\t\t%s\n", path);
	texture.type = type;
	texture.path = malloc(sizeof(char) * (strlen(path) + 1));
	strcpy(texture.path, path);
	texture.id = textureFromFile(path);
	*textures = texture;
}

static GLuint textureFromFile(char * path)
{
	GLuint textureID;
	glGenTextures(1, &textureID);
	int width, height, n;

	unsigned char * image = sx_stbi_load(path, &width, &height, &n, 0);
	glBindTexture(GL_TEXTURE_2D, textureID);
	if (n == 4) {
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, image);
	}
	else {
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, image);
	}
	sx_stbi_image_free(image);


	// Parameters
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D, 0);

	return textureID;
}

static size_t getIndicesCount(const struct aiMesh * aiMesh)
{
	unsigned int i;
	unsigned int num = aiMesh->mNumFaces;
	size_t count = 0;
	for (i = 0; i < num; i++)
	{
		struct aiFace face = aiMesh->mFaces[i];
		count += face.mNumIndices;
	}
	return count;
}


static void drawModel(Model * model, mat4x4 uMVMatrix, mat4x4 view, mat4x4 uPMatrix) {
	size_t i;
	model->program.Use(&model->program);
	GLuint uMVMatrixLoc = model->program.getUniformLocation(&model->program, "uMVMatrix");
	GLuint uPMatrixLoc = model->program.getUniformLocation(&model->program, "uPMatrix");
	glUniformMatrix4fv(uMVMatrixLoc, 1, GL_FALSE, (const GLfloat *)uMVMatrix);
	glUniformMatrix4fv(uPMatrixLoc, 1, GL_FALSE, (const GLfloat *)uPMatrix);

	vec4 lightAmbient = { 1.0f, 1.0f, 1.0f, 1.0f };
	vec4 lightDiffuse = { 1.0f, 1.0f, 1.0f, 1.0f };
	vec4 lightSpecular = { 1.0f, 1.0f, 1.0f, 1.0f };
	vec4 lightPosition = { 100.f, 100.f, 100.f, 1.0f };
	vec4 lightPositionV;
	mat4x4_mul_vec4(lightPositionV, view, lightPosition);

	glUniform4fv(model->program.getUniformLocation(&model->program, "light_ambient"), 1, (const GLfloat *) lightAmbient);
	glUniform4fv(model->program.getUniformLocation(&model->program, "light_diffuse"), 1, (const GLfloat *) lightDiffuse);
	glUniform4fv(model->program.getUniformLocation(&model->program, "light_specular"), 1, (const GLfloat *) lightSpecular);
	glUniform3fv(model->program.getUniformLocation(&model->program, "light_position"), 1, (const GLfloat *) lightPositionV);

	for (i = 0; i < model->numMeshes; i++) {
		model->meshes[i].Draw(&(model->meshes[i]), model);
	}
}

static void destroyModel(Model * model)
{
	for (unsigned int i = 0; i < model->numMeshes; i++) {
		free(model->meshes[i].vertices);
		free(model->meshes[i].indices);
		if (model->meshes[i].textures != NULL) {
			free(model->meshes[i].textures[0].path);
			free(model->meshes[i].textures[1].path);
			free(model->meshes[i].textures[2].path);
		}
		free(model->meshes[i].textures);
	}
	free(model->meshes);
}
/*------------------------------------------------*/
